export enum Difficulty {
	Easy,
	Medium,
	Hard,
	Last,
	Done
}

export type Case = number | undefined;

function shuffleArray<T>(array: Array<T>): void {
	for (let i = array.length - 1; i >= 0; i--) {
		const j = Math.floor(Math.random() * (i + 1));
		[array[i], array[j]] = [array[j], array[i]];
	}
}

function getNumberToKeepFromDifficulty(difficulty: Difficulty) {
	const rules = {
		0: 3,
		1: 2,
		2: 1,
		3: 8,
		4: 9
	};
	return rules[difficulty];
}

export class Board {
	grid: Case[][];
	annotations: number[][][];
	verticalSum: number[];
	horizontalSum: number[];
	lockedCells: boolean[][];

	constructor() {
		const numbers = Array.from({ length: 9 }, (_, i) => i + 1);
		shuffleArray(numbers);

		this.grid = [];
		this.annotations = [];
		this.verticalSum = new Array(3).fill(0);
		this.horizontalSum = [];
		this.lockedCells = [];

		for (let i = 0; i < 3; i++) {
			const row: number[] = [];
			for (let j = 0; j < 3; j++) {
				row[j] = numbers.pop() || 0;
				this.verticalSum[j] += row[j];
			}
			this.grid.push(row);
			this.annotations.push([[], [], []]);
			this.horizontalSum[i] = row.reduce((partialSum, a) => partialSum + a, 0);

			this.lockedCells[i] = [false, false, false];
		}
	}

	[Symbol.iterator]() {
		let index = 0;
		return {
			next: () => {
				if (index < 9) {
					const [row, col] = this.positionToRowCol(index++);
					return { value: this.grid[row][col], done: false };
				} else {
					return { done: true };
				}
			}
		};
	}

	isComplete(): boolean {
		for (const cell of this) {
			if (cell === undefined) return false;
		}
		return true;
	}

	isValid(): boolean {
		for (let i = 0; i < 3; i++) {
			const rowSum = this.grid[i].reduce((partialSum, a) => (partialSum || 0) + (a || 0), 0);
			const colSum = (this.grid[0][i] || 0) + (this.grid[1][i] || 0) + (this.grid[2][i] || 0);
			if (rowSum !== this.horizontalSum[i] || colSum !== this.verticalSum[i]) return false;
		}
		return true;
	}

	prepare(difficulty: Difficulty) {
		const numbers = Array.from({ length: 9 }, (_, i) => i);
		shuffleArray(numbers);
		const numberToKeep = getNumberToKeepFromDifficulty(difficulty);
		const posToHide = numbers.slice(numberToKeep, 9);

		posToHide.forEach((pos) => {
			this.setByPos(pos);
			// lock visible cells
			const [row, col] = this.positionToRowCol(pos);
			this.lockedCells[row][col] = true;
		});
	}

	reset() {
		for (let i = 0; i < 3; i++) {
			for (let j = 0; j < 3; j++) {
				if (this.lockedCells[i][j]) this.grid[i][j] = undefined;
				this.annotations[i][j] = [];
			}
		}
	}

	setByPos(pos: number, value: Case = undefined) {
		if (pos < 0 || pos > 8) throw new Error('position must be between 0 and 8!');
		if (value && (value < 1 || value > 9)) throw new Error('value must be between 1 and 9!');

		const [row, col] = this.positionToRowCol(pos);
		this.grid[row][col] = value;
	}

	positionToRowCol(position: number) {
		return [Math.floor(position / 3), position % 3];
	}

	getPlacedNumbers() {
		const res = [];
		for (const pos of this) {
			if (pos !== undefined) res.push(pos);
		}
		return res;
	}
}
